package com.sleepycat.je.log;

import java.io.File;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import latch.Latch;
import latch.LatchSupport;

import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.config.EnvironmentParams;
import com.sleepycat.je.config.IntConfigParam;
import com.sleepycat.je.dbi.DbConfigManager;
import com.sleepycat.je.dbi.EnvironmentImpl;

public privileged aspect FileHandleCaching perthis(newFileManager(*)){

	public static final IntConfigParam EnvironmentParams.LOG_FILE_CACHE_SIZE = new IntConfigParam(
			"je.log.fileCacheSize", new Integer(3), // min
			null, // max
			new Integer(100), // default
			false, // mutable
			"# The size of the file handle cache.");

	FileHandle around(long fileNum) throws LogException, DatabaseException :
		execution(FileHandle FileManager.getFileHandle(long)) && args(fileNum) {

		/* Check the file cache for this file. */
		Long fileId = new Long(fileNum);
		FileHandle fileHandle = null;

		/**
		 * Loop until we get an open FileHandle.
		 */
		while (true) {

			/*
			 * The file cache is intentionally not latched here so that it's not
			 * a bottleneck in the fast path. We check that the file handle that
			 * we get back is really still open after we latch it down below.
			 */
			fileHandle = fileCache.get(fileId);

			/* The file wasn't in the cache. */
			if (fileHandle == null) {
				fileCacheLatch.acquire();
				try {
					/* Check the file cache again under the latch. */
					fileHandle = fileCache.get(fileId);
					if (fileHandle == null) {

						fileHandle = proceed(fileNum);

						/* Put it into the cache. */
						fileCache.add(fileId, fileHandle);
					}
				} finally {
					fileCacheLatch.release();
				}
			}

			/* Get latch before returning */
			fileHandle.latch();
			/*
			 * We may have obtained this file handle outside the file cache
			 * latch, so we have to test that the handle is still valid. If it's
			 * not, then loop back and try again.
			 */
			if (fileHandle.getFile() == null) {
				fileHandle.release();
			} else {
				break;
			}
		}

		return fileHandle;
	}

	before() throws IOException, DatabaseException :
		execution(void FileManager.clear()){
		fileCacheLatch.acquire();
		try {
			fileCache.clear();
		} finally {
			fileCacheLatch.release();
		}
	}

	/*
	 * The file cache keeps N RandomAccessFile objects cached for file access.
	 * The cache consists of two parts: a Hashtable that doesn't require extra
	 * synchronization, for the most common access, and a linked list of files
	 * to support cache administration. Looking up a file from the hash table
	 * doesn't require extra latching, but adding or deleting a file does.
	 */
	private static class FileCache {
		private Map fileMap; // Long->file

		private LinkedList fileList; // list of file numbers

		private int fileCacheSize;

		FileCache(DbConfigManager configManager) throws DatabaseException {

			/*
			 * A fileMap maps the file number to FileHandles (RandomAccessFile,
			 * latch). The fileList is a list of Longs to determine which files
			 * to eject out of the file cache if it's too small.
			 */
			fileMap = new Hashtable();
			fileList = new LinkedList();
			fileCacheSize = configManager
					.getInt(EnvironmentParams.LOG_FILE_CACHE_SIZE);
		}

		private FileHandle get(Long fileId) {
			return (FileHandle) fileMap.get(fileId);
		}

		private void add(Long fileId, FileHandle fileHandle)
				throws DatabaseException {

			/*
			 * Does the cache have any room or do we have to evict? Hunt down
			 * the file list for an unused file. Note that the file cache might
			 * actually grow past the prescribed size if there is nothing
			 * evictable. Should we try to shrink the file cache? Presently if
			 * it grows, it doesn't shrink.
			 */
			if (fileList.size() >= fileCacheSize) {
				Iterator iter = fileList.iterator();
				while (iter.hasNext()) {
					Long evictId = (Long) iter.next();
					FileHandle evictTarget = (FileHandle) fileMap.get(evictId);

					/*
					 * Try to latch. If latchNoWait returns false, then another
					 * thread owns this latch. Note that a thread that's trying
					 * to get a new file handle should never already own the
					 * latch on another file handle, because these latches are
					 * meant to be short lived and only held over the i/o out of
					 * the file.
					 */
					if (evictTarget.latchNoWait()) {
						try {
							fileMap.remove(evictId);
							iter.remove();
							evictTarget.close();
						} catch (IOException e) {
							throw new DatabaseException(e);
						} finally {
							evictTarget.release();
						}
						break;
					}
				}
			}

			/*
			 * We've done our best to evict. Add the file the the cache now
			 * whether or not we did evict.
			 */
			fileList.add(fileId);
			fileMap.put(fileId, fileHandle);
		}

		/**
		 * Take any file handles corresponding to this file name out of the
		 * cache. A file handle could be there twice, in rd only and in r/w
		 * mode.
		 */
		private void remove(long fileNum) throws IOException, DatabaseException {

			Iterator iter = fileList.iterator();
			while (iter.hasNext()) {
				Long evictId = (Long) iter.next();
				if (evictId.longValue() == fileNum) {
					FileHandle evictTarget = (FileHandle) fileMap.get(evictId);
					try {
						evictTarget.latch();
						fileMap.remove(evictId);
						iter.remove();
						evictTarget.close();
					} finally {
						evictTarget.release();
					}
				}
			}
		}

		private void clear() throws IOException, DatabaseException {

			Iterator iter = fileMap.values().iterator();
			while (iter.hasNext()) {
				FileHandle fileHandle = (FileHandle) iter.next();
				try {
					fileHandle.latch();
					fileHandle.close();
					iter.remove();
				} finally {
					fileHandle.release();
				}
			}
			fileMap.clear();
			fileList.clear();
		}

		private Set getCacheKeys() {
			return fileMap.keySet();
		}
	}

	/* File cache */
	private FileCache fileCache;

	pointcut newFileManager(EnvironmentImpl env) : 
		execution(FileManager.new(EnvironmentImpl, File, boolean)) && 
		args(env,File,boolean);

	after(EnvironmentImpl env) throws DatabaseException :
		newFileManager(env)  {

		/* Cache of files. */
		DbConfigManager configManager = env.getConfigManager();
		fileCache = new FileCache(configManager);
		fileCacheLatch = LatchSupport.makeLatch(FileManager.class.getName()
				+ "_fileCache", env);
	}

	private Latch fileCacheLatch;

	/*
	 * Unit test support
	 */

	/*
	 * @return ids of files in cache
	 */
	Set getCacheKeys() {
		return fileCache.getCacheKeys();
	}

	/**
	 * Clear a file out of the file cache regardless of mode type.
	 */
	private void clearFileCache(long fileNum) throws IOException,
			DatabaseException {

		fileCacheLatch.acquire();
		try {
			fileCache.remove(fileNum);
		} finally {
			fileCacheLatch.release();
		}
	}

	before(long fileNum) throws IOException, DatabaseException:
			(execution(void FileManager.deleteFile(long)) ||
			execution(void FileManager.renameFile(long,String)))
			&& args(fileNum)
			{

		clearFileCache(fileNum);
	}
}
